// Microsoft Visual C++ generated resource script.
//
#include "resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include "afxres.h"

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// Japanese resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_JPN)
#ifdef _WIN32
LANGUAGE LANG_JAPANESE, SUBLANG_DEFAULT
#pragma code_page(932)
#endif //_WIN32


/////////////////////////////////////////////////////////////////////////////
//
// Version
//

#ifdef _WIN64
  #ifndef MOZC_RES_INTERNAL_NAME
    #if defined(GOOGLE_JAPANESE_INPUT_BUILD)
      #define MOZC_RES_INTERNAL_NAME "GoogleIMEJaInstallerHelper64"
    #elif defined(MOZC_BUILD)
      #define MOZC_RES_INTERNAL_NAME "mozc_installer_helper64"
    #endif
  #endif
  #ifndef MOZC_RES_ORIGINAL_FILENAME
    #if defined(GOOGLE_JAPANESE_INPUT_BUILD)
      #define MOZC_RES_ORIGINAL_FILENAME "GoogleIMEJaInstallerHelper64.dll"
    #elif defined(MOZC_BUILD)
      #define MOZC_RES_ORIGINAL_FILENAME "mozc_installer_helper64.dll"
    #endif
  #endif
#else
  #ifndef MOZC_RES_INTERNAL_NAME
    #if defined(GOOGLE_JAPANESE_INPUT_BUILD)
      #define MOZC_RES_INTERNAL_NAME "GoogleIMEJaInstallerHelper32"
    #elif defined(MOZC_BUILD)
      #define MOZC_RES_INTERNAL_NAME "mozc_installer_helper32"
    #endif
  #endif
  #ifndef MOZC_RES_ORIGINAL_FILENAME
    #if defined(GOOGLE_JAPANESE_INPUT_BUILD)
      #define MOZC_RES_ORIGINAL_FILENAME "GoogleIMEJaInstallerHelper32.dll"
    #elif defined(MOZC_BUILD)
      #define MOZC_RES_ORIGINAL_FILENAME "mozc_installer_helper32.dll"
    #endif
  #endif
#endif
#ifndef MOZC_RES_FILE_DESCRIPTION
  #if defined(GOOGLE_JAPANESE_INPUT_BUILD)
    #define MOZC_RES_FILE_DESCRIPTION "Google 日本語入力 インストーラー ヘルパー"
  #elif defined(MOZC_BUILD)
    #define MOZC_RES_FILE_DESCRIPTION "Mozc Installer Helper"
  #endif
#endif
#ifndef MOZC_RES_FILETYPE
  #define MOZC_RES_FILETYPE VFT_DLL
#endif


#ifdef MOZC_RES_USE_TEMPLATE
  #include "build_tools/mozc_win32_resource_template.rc"
#endif    // MOZC_RES_USE_TEMPLATE


/////////////////////////////////////////////////////////////////////////////
//
// String Table
//

STRINGTABLE LANGUAGE LANG_JAPANESE, SUBLANG_DEFAULT
BEGIN
#if defined(GOOGLE_JAPANESE_INPUT_BUILD)
  IDS_FORMAT_VERSION_INFO       "Google 日本語入力 インストーラー (バージョン %1!s!)"
  IDS_UNEXPECTED_ERROR          "未知のエラーが発生しました。"
  IDS_FORMAT_FUNCTION_AND_LINE  "%1!s! (%2!u!) でエラーが発生したため、インストールを中止しました。"
#elif defined(MOZC_BUILD)
  IDS_FORMAT_VERSION_INFO       "Mozc インストーラー (バージョン %1!s!)"
  IDS_UNEXPECTED_ERROR          "未知のエラーが発生しました。"
  IDS_FORMAT_FUNCTION_AND_LINE  "%1!s! (%2!u!) でエラーが発生したため、インストールを中止しました。"
#endif
END

STRINGTABLE LANGUAGE LANG_ENGLISH, SUBLANG_NEUTRAL
BEGIN
#if defined(GOOGLE_JAPANESE_INPUT_BUILD)
  IDS_FORMAT_VERSION_INFO       "Google Japanese Input Installer (Ver. %1!s!)"
  IDS_UNEXPECTED_ERROR          "An unknown error occurred."
  IDS_FORMAT_FUNCTION_AND_LINE  "The installation was stopped because of an error at %1!s! (%2!u!)"
#elif defined(MOZC_BUILD)
  IDS_FORMAT_VERSION_INFO       "Mozc Installer (Ver. %1!s!)"
  IDS_UNEXPECTED_ERROR          "An unknown error occurred."
  IDS_FORMAT_FUNCTION_AND_LINE  "The installation was stopped because of an error at %1!s! (%2!u!)"
#endif
END

#endif    // Japanese resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//


/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

