EXPORTS
        ImeConversionList
        ImeConfigure
        ImeDestroy
        ImeEscape
        ImeInquire
        ImeProcessKey
        ImeSelect
        ImeSetActiveContext
        ImeToAsciiEx
        NotifyIME
        ImeRegisterWord
        ImeUnregisterWord
        ImeGetRegisterWordStyle
        ImeEnumRegisterWord
        ImeSetCompositionString
        ImeGetImeMenuItems
