// Copyright 2010-2020, Google Inc.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
// copyright notice, this list of conditions and the following disclaimer
// in the documentation and/or other materials provided with the
// distribution.
//     * Neither the name of Google Inc. nor the names of its
// contributors may be used to endorse or promote products derived from
// this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "ios/ios_engine.h"

namespace mozc {

class SessionHandlerInterface {
 public:
  SessionHandlerInterface() = default;
  ~SessionHandlerInterface() = default;
};

namespace ios {

void IosEngine::InitMozc() {
  // Do nothing.
}

IosEngine::IosEngine(const std::string &data_file_path)
    : session_handler_(new SessionHandlerInterface), session_id_(0) {}

IosEngine::~IosEngine() = default;

bool IosEngine::SetMobileRequest(const std::string &keyboard_layout,
                                 commands::Command *command) {
  return true;
}

void FillMobileConfig(config::Config *config) {
  // Do nothing.
}

bool IosEngine::SetConfig(const config::Config &config,
                          commands::Command *command) {
  return true;
}

bool IosEngine::CreateSession(commands::Command *command) {
  return true;
}

bool IosEngine::DeleteSession(commands::Command *command) {
  return true;
}

bool IosEngine::ResetContext(commands::Command *command) {
  return true;
}

bool IosEngine::SendSpecialKey(commands::KeyEvent::SpecialKey special_key,
                               commands::Command *command) {
  return true;
}

bool IosEngine::SendKey(const std::string &character,
                        commands::Command *command) {
  return true;
}

bool IosEngine::SendSessionCommand(
    const commands::SessionCommand &session_command,
    commands::Command *command) {
  return true;
}

}  // namespace ios
}  // namespace mozc
