// Microsoft Visual C++ generated resource script.
//
#include "win32\base\display_name_resource.h"
#include "win32\tip\tip_resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include "afxres.h"

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// Japanese resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_JPN)
#ifdef _WIN32
LANGUAGE LANG_JAPANESE, SUBLANG_DEFAULT
#endif //_WIN32

/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
IDI_IMM32                 ICON       "data/images/win/product_icon_langbar.ico"
IDI_DIRECT                ICON       "data/images/win/ms_direct_input_a.ico"
IDI_HIRAGANA              ICON       "data/images/win/ms_hiragana_a.ico"
IDI_TOOL                  ICON       "data/images/win/tools_icon_a.ico"
IDI_TOOL_NT               ICON       "data/images/win/tools_icon.ico"
IDI_PROPERTY              ICON       "data/images/win/tools_properties_a.ico"
IDI_PROPERTY_NT           ICON       "data/images/win/tools_properties.ico"
IDI_DICTIONARY            ICON       "data/images/win/tools_dictionary_a.ico"
IDI_DICTIONARY_NT         ICON       "data/images/win/tools_dictionary.ico"
IDI_FULL_KATAKANA         ICON       "data/images/win/ms_katakana_a.ico"
IDI_HALF_ALPHANUMERIC     ICON       "data/images/win/ms_alpha_half_a.ico"
IDI_FULL_ALPHANUMERIC     ICON       "data/images/win/ms_alpha_a.ico"
IDI_HALF_KATAKANA         ICON       "data/images/win/ms_katakana_half_a.ico"
IDI_DISABLED              ICON       "data/images/win/ms_disabled_a.ico"
IDI_DISABLED_NT           ICON       "data/images/win/ms_disabled.ico"

#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE
BEGIN
    "tip_resource.h\0"
END

3 TEXTINCLUDE
BEGIN
    "\r\0"
END

2 TEXTINCLUDE
BEGIN
    "#include ""afxres.h""\r\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Version
//

#ifdef _WIN64
  #ifndef MOZC_RES_INTERNAL_NAME
    #if defined(GOOGLE_JAPANESE_INPUT_BUILD)
      #define MOZC_RES_INTERNAL_NAME "GoogleIMEJaTIP64"
    #elif defined(MOZC_BUILD)
      #define MOZC_RES_INTERNAL_NAME "mozc_ja_tip64"
    #endif
  #endif
  #ifndef MOZC_RES_ORIGINAL_FILENAME
    #if defined(GOOGLE_JAPANESE_INPUT_BUILD)
      #define MOZC_RES_ORIGINAL_FILENAME "GoogleIMEJaTIP64.dll"
    #elif defined(MOZC_BUILD)
      #define MOZC_RES_ORIGINAL_FILENAME "mozc_ja_tip64.dll"
    #endif
  #endif
#else
  #ifndef MOZC_RES_INTERNAL_NAME
    #if defined(GOOGLE_JAPANESE_INPUT_BUILD)
      #define MOZC_RES_INTERNAL_NAME "GoogleIMEJaTIP32"
    #elif defined(MOZC_BUILD)
      #define MOZC_RES_INTERNAL_NAME "mozc_ja_tip32"
    #endif
  #endif
  #ifndef MOZC_RES_ORIGINAL_FILENAME
    #if defined(GOOGLE_JAPANESE_INPUT_BUILD)
      #define MOZC_RES_ORIGINAL_FILENAME "GoogleIMEJaTIP32.dll"
    #elif defined(MOZC_BUILD)
      #define MOZC_RES_ORIGINAL_FILENAME "mozc_ja_tip32.dll"
    #endif
  #endif
#endif
#ifndef MOZC_RES_FILE_DESCRIPTION
  #if defined(GOOGLE_JAPANESE_INPUT_BUILD)
    #define MOZC_RES_FILE_DESCRIPTION "Google 日本語入力 TIP モジュール"
  #elif defined(MOZC_BUILD)
    #define MOZC_RES_FILE_DESCRIPTION "Mozc TIP Module"
  #endif
#endif
#ifndef MOZC_RES_FILETYPE
  #define MOZC_RES_FILETYPE VFT_DLL
#endif

#ifdef MOZC_RES_USE_TEMPLATE
  #include "build_tools/mozc_win32_resource_template.rc"
#endif    // MOZC_RES_USE_TEMPLATE

/////////////////////////////////////////////////////////////////////////////
//
// String Table
//

STRINGTABLE LANGUAGE LANG_JAPANESE, SUBLANG_DEFAULT
BEGIN
#if defined(GOOGLE_JAPANESE_INPUT_BUILD)
    IDS_IME_DISPLAYNAME     "Google 日本語入力"
    IDS_TEXTSERVICE_DISPLAYNAME_SYNONYM  "Google 日本語入力"
#else
    IDS_IME_DISPLAYNAME     "Mozc"
    IDS_TEXTSERVICE_DISPLAYNAME_SYNONYM  "Mozc"
#endif
    IDS_INPUTMODE           "入力モード"
    IDS_WIN8_TRAY_ITEM      "右クリックで IME のオプションを表示します"
    IDS_DIRECT              "直接入力"
    IDS_HIRAGANA            "ひらがな"
    IDS_TOOL                "ツール"
    IDS_PROPERTY            "プロパティ"
    IDS_DICTIONARY          "辞書ツール"
    IDS_WORD_REGISTER       "単語登録"
    IDS_HELP                "ヘルプ"
    IDS_CANCEL              "キャンセル"
#if defined(GOOGLE_JAPANESE_INPUT_BUILD)
    IDS_ABOUT               "Google 日本語入力について"
#else
    IDS_ABOUT               "Mozc について"
#endif
    IDS_FULL_KATAKANA       "全角カタカナ"
    IDS_HALF_ALPHANUMERIC   "半角英数"
    IDS_FULL_ALPHANUMERIC   "全角英数"
    IDS_HALF_KATAKANA       "半角カタカナ"
    IDS_RECONVERSION        "再変換"
    IDS_DISABLED            "IME が無効です"
#if defined(GOOGLE_JAPANESE_INPUT_BUILD)
    IDS_UNOBSERVABLE_SUGGEST_WINDOW  "Google 日本語入力 サジェストウィンドウ"
    IDS_OBSERVABLE_SUGGEST_WINDOW    "Google 日本語入力 サジェスト候補ウィンドウ"
    IDS_CANDIDATE_WINDOW             "Google 日本語入力 候補ウィンドウ"
    IDS_INDICATOR_WINDOW             "Google 日本語入力 インジケータ ウィンドウ"
#else
    IDS_UNOBSERVABLE_SUGGEST_WINDOW  "Mozc サジェストウィンドウ"
    IDS_OBSERVABLE_SUGGEST_WINDOW    "Mozc サジェスト候補ウィンドウ"
    IDS_CANDIDATE_WINDOW             "Mozc 候補ウィンドウ"
    IDS_INDICATOR_WINDOW             "Mozc インジケータ ウィンドウ"
#endif
END

STRINGTABLE LANGUAGE LANG_ENGLISH, SUBLANG_NEUTRAL
BEGIN
#if defined(GOOGLE_JAPANESE_INPUT_BUILD)
    IDS_IME_DISPLAYNAME     "Google Japanese Input"
    IDS_TEXTSERVICE_DISPLAYNAME_SYNONYM  "Google Japanese Input"
#else
    IDS_IME_DISPLAYNAME     "Mozc"
    IDS_TEXTSERVICE_DISPLAYNAME_SYNONYM  "Mozc"
#endif
    IDS_INPUTMODE           "Input Mode"
    IDS_WIN8_TRAY_ITEM      "Right click to show IME options"
    IDS_DIRECT              "Direct Input"
    IDS_HIRAGANA            "Hiragana"
    IDS_TOOL                "Tools"
    IDS_PROPERTY            "Properties"
    IDS_DICTIONARY          "Dictionary Tool"
    IDS_WORD_REGISTER       "Add Word"
    IDS_HELP                "Help"
    IDS_CANCEL              "Cancel"
#if defined(GOOGLE_JAPANESE_INPUT_BUILD)
    IDS_ABOUT               "About Google Japanese Input"
#else
    IDS_ABOUT               "About Mozc"
#endif
    IDS_FULL_KATAKANA       "Full-width Katakana"
    IDS_HALF_ALPHANUMERIC   "Half-width Alphanumeric"
    IDS_FULL_ALPHANUMERIC   "Full-width Alphanumeric"
    IDS_HALF_KATAKANA       "Half-width Katakana"
    IDS_RECONVERSION        "Reconversion"
    IDS_DISABLED            "IME is disabled"
#if defined(GOOGLE_JAPANESE_INPUT_BUILD)
    IDS_UNOBSERVABLE_SUGGEST_WINDOW  "Google Japanese Input Suggestion Window"
    IDS_OBSERVABLE_SUGGEST_WINDOW    "Google Japanese Input Suggestion Candidate Window"
    IDS_CANDIDATE_WINDOW             "Google Japanese Input Candidate Window"
#else
    IDS_UNOBSERVABLE_SUGGEST_WINDOW  "Mozc Suggestion Window"
    IDS_OBSERVABLE_SUGGEST_WINDOW    "Mozc Suggestion Candidate Window"
    IDS_CANDIDATE_WINDOW             "Mozc Candidate Window"
#endif
END

#endif    // Japanese resources
/////////////////////////////////////////////////////////////////////////////

#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//


/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED
