名詞	NOUN	*	名詞,一般,*,*,*,*
短縮よみ	ABBREVIATION	*	特殊,短縮よみ
サジェストのみ	 SUGGESTION_ONLY	*	特殊,サジェストのみ
固有名詞	PROPER_NOUN	*	名詞,固有名詞,一般,*,*,*
人名	PERSONAL_NAME	*	名詞,固有名詞,人名,一般,*,*
姓	FAMILY_NAME	*	名詞,固有名詞,人名,姓,*,*
名	FIRST_NAME	*	名詞,固有名詞,人名,名,*,*
組織	ORGANIZATION_NAME	*	名詞,固有名詞,組織,*,*,*
地名	PLACE_NAME	*	名詞,固有名詞,地域,一般,*,*
名詞サ変	SA_IRREGULAR_CONJUGATION_NOUN	*	名詞,サ変接続,*,*,*,*
名詞形動 ADJECTIVE_VERBAL_NOUN	*	名詞,形容動詞語幹,*,*,*,*
数	NUMBER	*	名詞,数,アラビア数字
アルファベット	ALPHABET	*	記号,アルファベット,*,*,*,*
記号	SYMBOL	*	記号,一般,*,*,*,*
顔文字	EMOTICON	*	記号,一般,*,*,*,*
副詞	ADVERB	*	副詞,一般,*,*,*,*
連体詞	PRENOUN_ADJECTIVAL	*	連体詞,*,*,*,*,*
接続詞	CONJUNCTION	*	接続詞,*,*,*,*,*
感動詞	INTERJECTION	*	感動詞,*,*,*,*,*
接頭語	PREFIX	*	接頭詞,名詞接続,*,*,*,*
助数詞	COUNTER_SUFFIX	*	名詞,接尾,助数詞,*,*,*
接尾一般	GENERIC_SUFFIX	*	名詞,接尾,一般,*,*,*
接尾人名	PERSON_NAME_SUFFIX	*	名詞,接尾,人名,*,*,*
接尾地名	PLACE_NAME_SUFFIX	*	名詞,接尾,地域,*,*,*
動詞ワ行五段	WA_GROUP1_VERB	五段・ワ行促音便	動詞,自立,*,*,五段・ワ行促音便,<cform>,*
動詞カ行五段	KA_GROUP1_VERB	五段・カ行イ音便	動詞,自立,*,*,五段動詞,<cform>,*
動詞サ行五段	SA_GROUP1_VERB	五段・サ行	動詞,自立,*,*,五段動詞,<cform>,*
動詞タ行五段	TA_GROUP1_VERB	五段・タ行	動詞,自立,*,*,五段動詞,<cform>,*
動詞ナ行五段	NA_GROUP1_VERB	五段・ナ行	動詞,自立,*,*,五段動詞,<cform>,*
動詞マ行五段	MA_GROUP1_VERB	五段・マ行	動詞,自立,*,*,五段動詞,<cform>,*
動詞ラ行五段	RA_GROUP1_VERB	五段・ラ行	動詞,自立,*,*,五段動詞,<cform>,*
動詞ガ行五段	GA_GROUP1_VERB	五段・ガ行	動詞,自立,*,*,五段動詞,<cform>,*
動詞バ行五段	BA_GROUP1_VERB	五段・バ行	動詞,自立,*,*,五段動詞,<cform>,*
動詞ハ行四段	HA_GROUP1_VERB	四段・ハ行	動詞,自立,*,*,五段動詞,<cform>,*
動詞一段	GROUP2_VERB	一段	動詞,自立,*,*,一段,<cform>,*
動詞カ変	KURU_GROUP3_VERB	カ変・来ル	動詞,自立,*,*,カ変・来ル,仮定形,*
動詞サ変	SURU_GROUP3_VERB	サ変・−スル	動詞,自立,*,*,サ変,<cform>,*
動詞ザ変	ZURU_GROUP3_VERB	サ変・−ズル	動詞,自立,*,*,サ変,<cform>,*
動詞ラ変	RU_GROUP3_VERB	ラ変	動詞,自立,*,*,ラ変,<cform>,*
形容詞	ADJECTIVE	形容詞・アウオ段	形容詞,自立,*,*,形容詞・アウオ段,<cform>,*
終助詞	SENTENCE_ENDING_PARTICLE	*	助詞,終助詞,*,*,*,*,ね
句読点	PUNCTUATION	*	記号,読点,*,*,*,*
独立語	FREE_STANDING_WORD	*	記号,一般,*,*,*,*
抑制単語	SUPPRESSION_WORD	*	記号,一般,*,*,*,*
