# 1st column: category of inflection
# 2st column: type of inflection
# 3rd column: suffix for value (* means NULL)
# 4th column: suffix for key  (* means NULL)
# 5th column: suffix for pronunciation  (* means NULL)
#
# currently, pronunciation is not used
#
# How to expand inflections:
# 1. Get the suffix of base form (基本形)
# 2. Remove the suffix of base form from the value (or key)
# 3. Add other suffixes, like suffix of 未然形 to the value made in step 2.
#
カ変・クル	基本形	くる	くる	くる
カ変・クル	未然形	こ	こ	こ
カ変・クル	未然ウ接続	こよ	こよ	こよ
カ変・クル	連用形	き	き	き
カ変・クル	仮定形	くれ	くれ	くれ
カ変・クル	命令ｙｏ	こよ	こよ	こよ
カ変・クル	命令ｉ	こい	こい	こい
カ変・クル	仮定縮約１	くりゃ	くりゃ	くりゃ
カ変・クル	体言接続特殊	くん	くん	くん
カ変・クル	体言接続特殊２	く	く	く
カ変・来ル	基本形	来る	くる	くる
カ変・来ル	未然形	来	こ	こ
カ変・来ル	未然ウ接続	来よ	こよ	こよ
カ変・来ル	連用形	来	き	き
カ変・来ル	仮定形	来れ	くれ	くれ
カ変・来ル	命令ｙｏ	来よ	こよ	こよ
カ変・来ル	命令ｉ	来い	こい	こい
カ変・来ル	仮定縮約１	来りゃ	くりゃ	くりゃ
カ変・来ル	体言接続特殊	来ん	くん	くん
カ変・来ル	体言接続特殊２	来	く	く
サ変・スル	基本形	する	する	する
サ変・スル	文語基本形	す	す	す
サ変・スル	未然形	し	し	し
サ変・スル	未然ウ接続	しよ	しよ	しよ
サ変・スル	未然ウ接続	しょ	しょ	しょ
サ変・スル	未然レル接続	さ	さ	さ
サ変・スル	未然ヌ接続	せ	せ	せ
サ変・スル	連用形	し	し	し
サ変・スル	仮定形	すれ	すれ	すれ
サ変・スル	命令ｙｏ	せよ	せよ	せよ
サ変・スル	命令ｒｏ	しろ	しろ	しろ
サ変・スル	命令ｉ	せい	せい	せい
サ変・スル	仮定縮約１	すりゃ	すりゃ	すりゃ
サ変・スル	体言接続特殊	すん	すん	すん
サ変・スル	体言接続特殊２	す	す	す
サ変・−スル	基本形	する	する	する
サ変・−スル	文語基本形	す	す	す
サ変・−スル	未然形	し	し	し
サ変・−スル	未然ウ接続	しよ	しよ	しよ
サ変・−スル	未然ウ接続	しょ	しょ	しょ
サ変・−スル	未然レル接続	せ	せ	せ
サ変・−スル	仮定形	すれ	すれ	すれ
サ変・−スル	命令ｙｏ	せよ	せよ	せよ
サ変・−スル	命令ｒｏ	しろ	しろ	しろ
サ変・−スル	仮定縮約１	すりゃ	すりゃ	すりゃ
サ変・−ズル	基本形	ずる	ずる	ずる
サ変・−ズル	文語基本形	ず	ず	ず
サ変・−ズル	未然形	ぜ	ぜ	ぜ
サ変・−ズル	未然ウ接続	ぜよ	ぜよ	ぜよ
サ変・−ズル	仮定形	ずれ	ずれ	ずれ
サ変・−ズル	命令ｙｏ	ぜよ	ぜよ	ぜよ
サ変・−ズル	仮定縮約１	ずりゃ	ずりゃ	ずりゃ
一段	基本形	る	る	る
一段	未然形	*	*	*
一段	未然ウ接続	よ	よ	よ
一段	連用形	*	*	*
一段	仮定形	れ	れ	れ
一段	命令ｙｏ	よ	よ	よ
一段	命令ｒｏ	ろ	ろ	ろ
一段	仮定縮約１	りゃ	りゃ	りゃ
一段	体言接続特殊	ん	ん	ん
一段・病メル	基本形	る	る	る
五段・カ行イ音便	基本形	く	く	く
五段・カ行イ音便	未然形	か	か	か
五段・カ行イ音便	未然ウ接続	こ	こ	こ
五段・カ行イ音便	連用形	き	き	き
五段・カ行イ音便	連用タ接続	い	い	い
五段・カ行イ音便	仮定形	け	け	け
五段・カ行イ音便	命令ｅ	け	け	け
五段・カ行イ音便	仮定縮約１	きゃ	きゃ	きゃ
五段・カ行促音便	基本形	く	く	く
五段・カ行促音便	未然形	か	か	か
五段・カ行促音便	未然ウ接続	こ	こ	こ
五段・カ行促音便	連用形	き	き	き
五段・カ行促音便	連用タ接続	っ	っ	っ
五段・カ行促音便	仮定形	け	け	け
五段・カ行促音便	命令ｅ	け	け	け
五段・カ行促音便	仮定縮約１	きゃ	きゃ	きゃ
五段・カ行促音便ユク	基本形	く	く	く
五段・カ行促音便ユク	未然形	か	か	か
五段・カ行促音便ユク	未然ウ接続	こ	こ	こ
五段・カ行促音便ユク	連用形	き	き	き
五段・カ行促音便ユク	仮定形	け	け	け
五段・カ行促音便ユク	命令ｅ	け	け	け
五段・カ行促音便ユク	仮定縮約１	きゃ	きゃ	きゃ
五段・ガ行	基本形	ぐ	ぐ	ぐ
五段・ガ行	未然形	が	が	が
五段・ガ行	未然ウ接続	ご	ご	ご
五段・ガ行	連用形	ぎ	ぎ	ぎ
五段・ガ行	連用タ接続	い	い	い
五段・ガ行	仮定形	げ	げ	げ
五段・ガ行	命令ｅ	げ	げ	げ
五段・ガ行	仮定縮約１	ぎゃ	ぎゃ	ぎゃ
五段・サ行	基本形	す	す	す
五段・サ行	未然形	さ	さ	さ
五段・サ行	未然ウ接続	そ	そ	そ
五段・サ行	連用形	し	し	し
五段・サ行	仮定形	せ	せ	せ
五段・サ行	命令ｅ	せ	せ	せ
五段・サ行	仮定縮約１	しゃ	しゃ	しゃ
五段・タ行	基本形	つ	つ	つ
五段・タ行	未然形	た	た	た
五段・タ行	未然ウ接続	と	と	と
五段・タ行	連用形	ち	ち	ち
五段・タ行	連用タ接続	っ	っ	っ
五段・タ行	仮定形	て	て	て
五段・タ行	命令ｅ	て	て	て
五段・タ行	仮定縮約１	ちゃ	ちゃ	ちゃ
五段・ナ行	基本形	ぬ	ぬ	ぬ
五段・ナ行	未然形	な	な	な
五段・ナ行	未然ウ接続	の	の	の
五段・ナ行	連用形	に	に	に
五段・ナ行	連用タ接続	ん	ん	ん
五段・ナ行	仮定形	ね	ね	ね
五段・ナ行	命令ｅ	ね	ね	ね
五段・ナ行	仮定縮約１	にゃ	にゃ	にゃ
五段・バ行	基本形	ぶ	ぶ	ぶ
五段・バ行	未然形	ば	ば	ば
五段・バ行	未然ウ接続	ぼ	ぼ	ぼ
五段・バ行	連用形	び	び	び
五段・バ行	連用タ接続	ん	ん	ん
五段・バ行	仮定形	べ	べ	べ
五段・バ行	命令ｅ	べ	べ	べ
五段・バ行	仮定縮約１	びゃ	びゃ	びゃ
五段・マ行	基本形	む	む	む
五段・マ行	未然形	ま	ま	ま
五段・マ行	未然ウ接続	も	も	も
五段・マ行	連用形	み	み	み
五段・マ行	連用タ接続	ん	ん	ん
五段・マ行	仮定形	め	め	め
五段・マ行	命令ｅ	め	め	め
五段・マ行	仮定縮約１	みゃ	みゃ	みゃ
五段・ラ行	基本形	る	る	る
五段・ラ行	未然形	ら	ら	ら
五段・ラ行	未然特殊	ん	ん	ん
五段・ラ行	未然ウ接続	ろ	ろ	ろ
五段・ラ行	連用形	り	り	り
五段・ラ行	連用タ接続	っ	っ	っ
五段・ラ行	仮定形	れ	れ	れ
五段・ラ行	命令ｅ	れ	れ	れ
五段・ラ行	仮定縮約１	りゃ	りゃ	りゃ
五段・ラ行	体言接続特殊	ん	ん	ん
# 五段・ラ行	体言接続特殊２	*	*	*
五段・ラ行アル	基本形	る	る	る
五段・ラ行アル	未然形	ら	ら	ら
五段・ラ行アル	未然ウ接続	ろ	ろ	ろ
五段・ラ行アル	連用形	り	り	り
五段・ラ行アル	連用タ接続	っ	っ	っ
五段・ラ行アル	仮定形	れ	れ	れ
五段・ラ行アル	命令ｅ	れ	れ	れ
五段・ラ行アル	仮定縮約１	りゃ	りゃ	りゃ
五段・ラ行アル	体言接続特殊	ん	ん	ん
五段・ラ行特殊	基本形	る	る	る
五段・ラ行特殊	未然形	ら	ら	ら
五段・ラ行特殊	未然特殊	ん	ん	ん
五段・ラ行特殊	未然ウ接続	ろ	ろ	ろ
五段・ラ行特殊	連用形	い	い	い
五段・ラ行特殊	連用タ接続	っ	っ	っ
五段・ラ行特殊	仮定形	れ	れ	れ
五段・ラ行特殊	命令ｅ	れ	れ	れ
五段・ラ行特殊	命令ｉ	い	い	い
五段・ラ行特殊	仮定縮約１	りゃ	りゃ	りゃ
五段・ワ行ウ音便	基本形	う	う	う
五段・ワ行ウ音便	未然形	わ	わ	わ
五段・ワ行ウ音便	未然ウ接続	お	お	お
五段・ワ行ウ音便	連用形	い	い	い
五段・ワ行ウ音便	連用タ接続	う	う	う
五段・ワ行ウ音便	仮定形	え	え	え
五段・ワ行ウ音便	命令ｅ	え	え	え
五段・ワ行促音便	基本形	う	う	う
五段・ワ行促音便	未然形	わ	わ	わ
五段・ワ行促音便	未然ウ接続	お	お	お
五段・ワ行促音便	連用形	い	い	い
五段・ワ行促音便	連用タ接続	っ	っ	っ
五段・ワ行促音便	仮定形	え	え	え
五段・ワ行促音便	命令ｅ	え	え	え
四段・カ行	基本形	く	く	く
四段・カ行	未然形	か	か	か
四段・カ行	連用形	き	き	き
四段・カ行	仮定形	け	け	け
四段・カ行	命令ｅ	け	け	け
四段・ガ行	基本形	ぐ	ぐ	ぐ
四段・ガ行	未然形	が	が	が
四段・ガ行	連用形	ぎ	ぎ	ぎ
四段・ガ行	仮定形	げ	げ	げ
四段・ガ行	命令ｅ	げ	げ	げ
四段・サ行	基本形	す	す	す
四段・サ行	未然形	さ	さ	さ
四段・サ行	連用形	し	し	し
四段・サ行	仮定形	せ	せ	せ
四段・サ行	命令ｅ	せ	せ	せ
四段・タ行	基本形	つ	つ	つ
四段・タ行	未然形	た	た	た
四段・タ行	連用形	ち	ち	ち
四段・タ行	仮定形	て	て	て
四段・タ行	命令ｅ	て	て	て
四段・バ行	基本形	ぶ	ぶ	ぶ
四段・バ行	未然形	ば	ば	ば
四段・バ行	連用形	び	び	び
四段・バ行	仮定形	べ	べ	べ
四段・バ行	命令ｅ	べ	べ	べ
四段・マ行	基本形	む	む	む
四段・マ行	未然形	ま	ま	ま
四段・マ行	連用形	み	み	み
四段・マ行	仮定形	め	め	め
四段・マ行	命令ｅ	め	め	め
四段・ラ行	基本形	る	る	る
四段・ラ行	未然形	ら	ら	ら
四段・ラ行	連用形	り	り	り
四段・ラ行	仮定形	れ	れ	れ
四段・ラ行	命令ｅ	れ	れ	れ
四段・ハ行	基本形	ふ	ふ	う
四段・ハ行	未然形	は	は	わ
四段・ハ行	連用形	ひ	ひ	い
四段・ハ行	仮定形	へ	へ	え
四段・ハ行	命令ｅ	へ	へ	え
ラ変	基本形	り	り	り
ラ変	未然形	ら	ら	ら
ラ変	連用形	り	り	り
ラ変	体言接続	る	る	る
ラ変	仮定形	れ	れ	れ
ラ変	命令ｅ	れ	れ	れ
上二・ダ行	基本形	づ	づ	ず
上二・ダ行	現代基本形	ず	ず	ず
上二・ダ行	未然形	ぢ	ぢ	じ
上二・ダ行	連用形	ぢ	ぢ	じ
上二・ダ行	体言接続	づる	づる	ずる
上二・ダ行	体言接続	ずる	ずる	ずる
上二・ダ行	仮定形	づれ	づれ	ずれ
上二・ダ行	仮定形	ずれ	ずれ	ずれ
上二・ダ行	命令ｙｏ	ぢよ	ぢよ	じよ
上二・ハ行	基本形	ふ	ふ	う
上二・ハ行	未然形	ひ	ひ	い
上二・ハ行	連用形	ひ	ひ	い
上二・ハ行	体言接続	ふる	ふる	うる
上二・ハ行	仮定形	ふれ	ふれ	うれ
上二・ハ行	命令ｙｏ	ひよ	ひよ	いよ
下二・ア行	基本形	う	う	う
下二・ア行	未然形	え	え	え
下二・ア行	連用形	え	え	え
下二・ア行	体言接続	うる	うる	うる
下二・ア行	仮定形	うれ	うれ	うれ
下二・ア行	命令ｙｏ	えよ	えよ	えよ
下二・カ行	基本形	く	く	く
下二・カ行	未然形	け	け	け
下二・カ行	連用形	け	け	け
下二・カ行	体言接続	くる	くる	くる
下二・カ行	仮定形	くれ	くれ	くれ
下二・カ行	命令ｙｏ	けよ	けよ	けよ
下二・ガ行	基本形	ぐ	ぐ	ぐ
下二・ガ行	未然形	げ	げ	げ
下二・ガ行	連用形	げ	げ	げ
下二・ガ行	体言接続	ぐる	ぐる	ぐる
下二・ガ行	仮定形	ぐれ	ぐれ	ぐれ
下二・ガ行	命令ｙｏ	げよ	げよ	げよ
下二・サ行	基本形	す	す	す
下二・サ行	未然形	せ	せ	せ
下二・サ行	連用形	せ	せ	せ
下二・サ行	体言接続	する	する	する
下二・サ行	仮定形	すれ	すれ	すれ
下二・サ行	命令ｙｏ	せよ	せよ	せよ
下二・ザ行	基本形	ず	ず	ず
下二・ザ行	未然形	ぜ	ぜ	ぜ
下二・ザ行	連用形	ぜ	ぜ	ぜ
下二・ザ行	体言接続	ずる	ずる	ずる
下二・ザ行	仮定形	ずれ	ずれ	ずれ
下二・ザ行	命令ｙｏ	ぜよ	ぜよ	ぜよ
下二・タ行	基本形	つ	つ	つ
下二・タ行	未然形	て	て	て
下二・タ行	連用形	て	て	て
下二・タ行	体言接続	つる	つる	つる
下二・タ行	仮定形	つれ	つれ	つれ
下二・タ行	命令ｙｏ	てよ	てよ	てよ
下二・ダ行	基本形	づ	づ	ず
下二・ダ行	未然形	で	で	で
下二・ダ行	連用形	で	で	で
下二・ダ行	体言接続	づる	づる	ずる
下二・ダ行	仮定形	づれ	づれ	ずれ
下二・ダ行	命令ｙｏ	でよ	でよ	でよ
下二・ナ行	基本形	ぬ	ぬ	ぬ
下二・ナ行	未然形	ね	ね	ね
下二・ナ行	連用形	ね	ね	ね
下二・ナ行	体言接続	ぬる	ぬる	ぬる
下二・ナ行	仮定形	ぬれ	ぬれ	ぬれ
下二・ナ行	命令ｙｏ	ねよ	ねよ	ねよ
下二・ハ行	基本形	ふ	ふ	う
下二・ハ行	未然形	へ	へ	え
下二・ハ行	連用形	へ	へ	え
下二・ハ行	体言接続	ふる	ふる	うる
下二・ハ行	仮定形	ふれ	ふれ	うれ
下二・ハ行	命令ｙｏ	へよ	へよ	えよ
下二・バ行	基本形	ぶ	ぶ	ぶ
下二・バ行	未然形	べ	べ	べ
下二・バ行	連用形	べ	べ	べ
下二・バ行	体言接続	ぶる	ぶる	ぶる
下二・バ行	仮定形	ぶれ	ぶれ	ぶれ
下二・バ行	命令ｙｏ	べよ	べよ	べよ
下二・マ行	基本形	む	む	む
下二・マ行	未然形	め	め	め
下二・マ行	連用形	め	め	め
下二・マ行	体言接続	むる	むる	むる
下二・マ行	仮定形	むれ	むれ	むれ
下二・マ行	命令ｙｏ	めよ	めよ	めよ
下二・ヤ行	基本形	ゆ	ゆ	ゆ
下二・ヤ行	未然形	え	え	え
下二・ヤ行	連用形	え	え	え
下二・ヤ行	体言接続	ゆる	ゆる	ゆる
下二・ヤ行	仮定形	ゆれ	ゆれ	ゆれ
下二・ヤ行	命令ｙｏ	えよ	えよ	えよ
下二・ラ行	基本形	る	る	る
下二・ラ行	未然形	れ	れ	れ
下二・ラ行	連用形	れ	れ	れ
下二・ラ行	体言接続	るる	るる	るる
下二・ラ行	仮定形	るれ	るれ	るれ
下二・ラ行	命令ｙｏ	れよ	れよ	れよ
下二・ワ行	基本形	う	う	う
下二・ワ行	未然形	ゑ	ゑ	え
下二・ワ行	連用形	ゑ	ゑ	え
下二・ワ行	体言接続	うる	うる	うる
下二・ワ行	仮定形	うれ	うれ	うれ
下二・ワ行	命令ｙｏ	ゑよ	ゑよ	えよ
下二・得	基本形	得	う	う
下二・得	未然形	得	え	え
下二・得	未然ウ接続	得よ	えよ	えよ
下二・得	連用形	得	え	え
下二・得	体言接続	得る	うる	うる
下二・得	仮定形	得れ	うれ	うれ
下二・得	命令ｙｏ	得よ	えよ	えよ
一段・クレル	基本形	れる	れる	れる
一段・クレル	未然形	れ	れ	れ
一段・クレル	未然特殊	ん	ん	ん
一段・クレル	未然ウ接続	れよ	れよ	れよ
一段・クレル	連用形	れ	れ	れ
一段・クレル	仮定形	れれ	れれ	れれ
一段・クレル	命令ｅ	れ	れ	れ
一段・クレル	命令ｙｏ	れよ	れよ	れよ
一段・クレル	命令ｒｏ	れろ	れろ	れろ
一段・クレル	仮定縮約１	れりゃ	れりゃ	れりゃ
一段・ル	基本形	る	る	る
一段・ル	仮定形	れ	れ	れ
一段・ル	命令ｒｏ	ろ	ろ	ろ
一段・ル	仮定縮約１	りゃ	りゃ	りゃ
形容詞・アウオ段	基本形	い	い	い
# 形容詞・アウオ段	文語基本形	し	し	し
形容詞・アウオ段	未然ヌ接続	から	から	から
形容詞・アウオ段	未然ウ接続	かろ	かろ	かろ
形容詞・アウオ段	連用タ接続	かっ	かっ	かっ
形容詞・アウオ段	連用テ接続	く	く	く
形容詞・アウオ段	連用テ接続	くっ	くっ	くっ
形容詞・アウオ段	連用ゴザイ接続	う	う	ー
形容詞・アウオ段	連用ゴザイ接続	ぅ	ぅ	ー
形容詞・アウオ段	体言接続	き	き	き
形容詞・アウオ段	仮定形	けれ	けれ	けれ
形容詞・アウオ段	命令ｅ	かれ	かれ	かれ
形容詞・アウオ段	仮定縮約１	けりゃ	けりゃ	けりゃ
形容詞・アウオ段	仮定縮約２	きゃ	きゃ	きゃ
形容詞・アウオ段	ガル接続	*	*	*
形容詞・イ段	基本形	い	い	い
形容詞・イ段	文語基本形	*	*	*
形容詞・イ段	未然ヌ接続	から	から	から
形容詞・イ段	未然ウ接続	かろ	かろ	かろ
形容詞・イ段	連用タ接続	かっ	かっ	かっ
形容詞・イ段	連用テ接続	く	く	く
形容詞・イ段	連用テ接続	くっ	くっ	くっ
形容詞・イ段	連用ゴザイ接続	ゅう	ゅう	ゅー
形容詞・イ段	連用ゴザイ接続	ゅぅ	ゅぅ	ゅー
形容詞・イ段	体言接続	き	き	き
形容詞・イ段	仮定形	けれ	けれ	けれ
形容詞・イ段	命令ｅ	かれ	かれ	かれ
形容詞・イ段	仮定縮約１	けりゃ	けりゃ	けりゃ
形容詞・イ段	仮定縮約２	きゃ	きゃ	きゃ
形容詞・イ段	ガル接続	*	*	*
特殊・ナイ	基本形	ない	ない	ない
特殊・ナイ	音便基本形	ねえ	ねえ	ねー
特殊・ナイ	音便基本形	ねぇ	ねぇ	ねー
特殊・ナイ	文語基本形	なし	なし	なし
特殊・ナイ	未然ヌ接続	なから	なから	なから
特殊・ナイ	未然ウ接続	なかろ	なかろ	なかろ
特殊・ナイ	連用タ接続	なかっ	なかっ	なかっ
特殊・ナイ	連用タ接続	なかつ	なかつ	なかっ
特殊・ナイ	連用テ接続	なく	なく	なく
特殊・ナイ	連用テ接続	なくっ	なくっ	なくっ
特殊・ナイ	連用デ接続	ない	ない	ない
特殊・ナイ	連用ゴザイ接続	のう	のう	のー
特殊・ナイ	体言接続	なき	なき	なき
特殊・ナイ	仮定形	なけれ	なけれ	なけれ
特殊・ナイ	命令ｅ	なかれ	なかれ	なかれ
特殊・ナイ	仮定縮約１	なけりゃ	なけりゃ	なけりゃ
特殊・ナイ	仮定縮約２	なきゃ	なきゃ	なきゃ
特殊・ナイ	ガル接続	な	な	な
特殊・タイ	基本形	たい	たい	たい
特殊・タイ	音便基本形	てえ	てえ	てー
特殊・タイ	音便基本形	てぇ	てぇ	てー
特殊・タイ	文語基本形	たし	たし	たし
特殊・タイ	未然ヌ接続	たから	たから	たから
特殊・タイ	未然ウ接続	たかろ	たかろ	たかろ
特殊・タイ	連用タ接続	たかっ	たかっ	たかっ
特殊・タイ	連用タ接続	たかつ	たかつ	たかっ
特殊・タイ	連用テ接続	たく	たく	たく
特殊・タイ	連用テ接続	たくっ	たくっ	たくっ
特殊・タイ	連用ゴザイ接続	とう	とう	とー
特殊・タイ	体言接続	たき	たき	たき
特殊・タイ	仮定形	たけれ	たけれ	たけれ
特殊・タイ	仮定縮約１	たけりゃ	たけりゃ	たけりゃ
特殊・タイ	仮定縮約２	たきゃ	たきゃ	たきゃ
特殊・タイ	ガル接続	た	た	た
特殊・タ	基本形	*	*	*
特殊・タ	未然形	ろ	ろ	ろ
特殊・タ	仮定形	ら	ら	ら
特殊・ダ	基本形	だ	だ	だ
特殊・ダ	未然形	だろ	だろ	だろ
特殊・ダ	未然形	だら	だら	だら
特殊・ダ	連用形	で	で	で
特殊・ダ	連用タ接続	だっ	だっ	だっ
特殊・ダ	連用用言接続	に	に	に
特殊・ダ	体言接続	な	な	な
特殊・ダ	仮定形	なら	なら	なら
特殊・ダ	命令ｅ	なれ	なれ	なれ
特殊・デス	基本形	す	す	す
特殊・デス	未然形	しょ	しょ	しょ
特殊・デス	連用形	し	し	し
特殊・ジャ	基本形	*	*	*
特殊・ジャ	未然形	ろ	ろ	ろ
特殊・ジャ	連用形	っ	っ	っ
特殊・マス	基本形	す	す	す
特殊・マス	未然形	せ	せ	せ
特殊・マス	未然ウ接続	しょ	しょ	しょ
特殊・マス	連用形	し	し	し
特殊・マス	仮定形	すれ	すれ	すれ
特殊・マス	命令ｅ	せ	せ	せ
特殊・マス	命令ｉ	し	し	し
特殊・ヌ	基本形	ぬ	ぬ	ぬ
特殊・ヌ	文語基本形	ず	ず	ず
特殊・ヌ	連用形	ざり	ざり	ざり
特殊・ヌ	連用ニ接続	ず	ず	ず
特殊・ヌ	体言接続	ざる	ざる	ざる
特殊・ヌ	仮定形	ね	ね	ね
特殊・ヌ	仮定形	ざれ	ざれ	ざれ
特殊・ヌ	仮定形	ずん	ずん	ずん
文語・ベシ	基本形	し	し	し
文語・ベシ	未然形	から	から	から
文語・ベシ	連用形	く	く	く
文語・ベシ	体言接続	き	き	き
文語・ベシ	仮定形	けれ	けれ	けれ
文語・ゴトシ	基本形	し	し	し
文語・ゴトシ	連用形	く	く	く
文語・ゴトシ	体言接続	き	き	き
文語・ナリ	基本形	り	り	り
文語・ナリ	未然形	ら	ら	ら
文語・ナリ	体言接続	る	る	る
文語・ナリ	仮定形	れ	れ	れ
文語・ナリ	命令ｅ	れ	れ	れ
文語・マジ	基本形	*	*	*
文語・マジ	連用形	く	く	く
文語・マジ	体言接続	き	き	き
文語・マジ	仮定形	けれ	けれ	けれ
文語・シム	基本形	む	む	む
文語・シム	未然形	め	め	め
文語・シム	連用形	め	め	め
文語・シム	体言接続	むる	むる	むる
文語・シム	仮定形	むれ	むれ	むれ
文語・シム	命令ｅ	むれ	むれ	むれ
文語・シム	命令ｙｏ	めよ	めよ	めよ
文語・キ	基本形	き	き	き
文語・キ	体言接続	し	し	し
文語・キ	命令ｅ	しか	しか	しか
文語・ケリ	基本形	り	り	り
文語・ケリ	体言接続	る	る	る
文語・リ	基本形	り	り	り
文語・リ	体言接続	る	る	る
文語・ル	基本形	る	る	る
文語・ル	未然形	れ	れ	れ
文語・ル	連用形	れ	れ	れ
文語・ル	体言接続	るる	るる	るる
文語・ル	仮定形	るれ	るれ	るれ
文語・ル	命令ｅ	るれ	るれ	るれ
文語・ル	命令ｙｏ	れよ	れよ	れよ
不変化型	基本形	*	*	*
形容詞・イイ	基本形	い	い	い
形容詞・イイ	基本形-促音便	いっ	いっ	いっ
形容詞・イイ	基本形-促音便	いいっ	いいっ	いいっ
特殊・ドス	基本形	す	す	す
一段・得ル	基本形	る	る	る
一段・得ル	仮定形	れ	れ	れ
特殊・ヤ	基本形	*	*	*
特殊・ヤ	未然形	ろ	ろ	ろ
特殊・ヤ	連用形	っ	っ	っ
