/////////////////////////////////////////////////////////////////////////////
//
// Version
//
#ifndef MOZC_RES_COMPANY_NAME
  #if defined(GOOGLE_JAPANESE_INPUT_BUILD)
    #define MOZC_RES_COMPANY_NAME "Google Inc."
  #elif defined(MOZC_BUILD)
    #define MOZC_RES_COMPANY_NAME "Google Inc."
  #endif
#endif
#ifndef MOZC_RES_FILE_DESCRIPTION
  #define MOZC_RES_FILE_DESCRIPTION ""
#endif
#ifndef MOZC_RES_INTERNAL_NAME
  #define MOZC_RES_INTERNAL_NAME ""
#endif
#ifndef MOZC_RES_LEGAL_COPYRIGHT
  #if defined(GOOGLE_JAPANESE_INPUT_BUILD)
    #define MOZC_RES_LEGAL_COPYRIGHT "Copyright (C) 2012 Google Inc. All Rights Reserved."
  #elif defined(MOZC_BUILD)
    #define MOZC_RES_LEGAL_COPYRIGHT "Copyright (C) 2012 Google Inc. All Rights Reserved."
  #endif
#endif
#ifndef MOZC_RES_ORIGINAL_FILENAME
  #define MOZC_RES_ORIGINAL_FILENAME ""
#endif
#ifndef MOZC_RES_PRODUCT_NAME
  #if defined(GOOGLE_JAPANESE_INPUT_BUILD)
    #define MOZC_RES_PRODUCT_NAME "Google 日本語入力"
  #elif defined(MOZC_BUILD)
    #define MOZC_RES_PRODUCT_NAME "Mozc"
  #endif
#endif
#ifndef MOZC_RES_FILE_VERSION_NUMBER
  #define MOZC_RES_FILE_VERSION_NUMBER MOZC_RES_VERSION_NUMBER
#endif
#ifndef MOZC_RES_PRODUCT_VERSION_NUMBER
  #define MOZC_RES_PRODUCT_VERSION_NUMBER MOZC_RES_VERSION_NUMBER
#endif
#ifndef MOZC_RES_FILE_VERSION_STRING
  #define MOZC_RES_FILE_VERSION_STRING MOZC_RES_SPECIFIC_VERSION_STRING
#endif
#ifndef MOZC_RES_PRODUCT_VERSION_STRING
  #define MOZC_RES_PRODUCT_VERSION_STRING MOZC_RES_VERSION_STRING
#endif
#ifndef MOZC_RES_FILEFLAGSMASK
  #define MOZC_RES_FILEFLAGSMASK (VS_FF_DEBUG | VS_FF_PRERELEASE | VS_FF_PATCHED | VS_FF_INFOINFERRED)
#endif
#ifndef MOZC_RES_FILEFLAGS
  #ifdef _DEBUG
    #define MOZC_RES_FILEFLAGS VS_FF_DEBUG
  #else
    #define MOZC_RES_FILEFLAGS 0x0L
  #endif
#endif
#ifndef MOZC_RES_FILEOS
  #define MOZC_RES_FILEOS VOS__WINDOWS32
#endif
#ifndef MOZC_RES_FILETYPE
  #define MOZC_RES_FILETYPE VFT_UNKNOWN
#endif
#ifndef MOZC_RES_FILESUBTYPE
  #define MOZC_RES_FILESUBTYPE VFT2_UNKNOWN
#endif

VS_VERSION_INFO VERSIONINFO
FILEVERSION MOZC_RES_FILE_VERSION_NUMBER
PRODUCTVERSION MOZC_RES_PRODUCT_VERSION_NUMBER
FILEFLAGSMASK MOZC_RES_FILEFLAGSMASK
FILEFLAGS MOZC_RES_FILEFLAGS
FILEOS MOZC_RES_FILEOS
FILETYPE MOZC_RES_FILETYPE
FILESUBTYPE MOZC_RES_FILESUBTYPE
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "041104b0"
        BEGIN
            VALUE "CompanyName", MOZC_RES_COMPANY_NAME
            VALUE "FileDescription", MOZC_RES_FILE_DESCRIPTION
            VALUE "FileVersion", MOZC_RES_FILE_VERSION_STRING
            VALUE "InternalName", MOZC_RES_INTERNAL_NAME
            VALUE "LegalCopyright", MOZC_RES_LEGAL_COPYRIGHT
            VALUE "OriginalFilename", MOZC_RES_ORIGINAL_FILENAME
            VALUE "ProductName", MOZC_RES_PRODUCT_NAME
            VALUE "ProductVersion", MOZC_RES_PRODUCT_VERSION_STRING
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x411, 1200
    END
END

