# This file defines mapping between POS of other IME's and Mozc. This file
# is read by gen_pos_map.py to produce pos_map.h. Mozc POS should be
# one of those defined in user_pos.def
#
# Each line defines one mapping rule to convert Source POS to
# Mozc's. Three types of format are allowed.
#
# Note that fields should be tab-separated and no leading or trailing
# white space is allowed.
#
# 1.
# <SOURCE POS>
# Words of <SOURCE POS> are discarded in conversion.
#
# 2.
# <SOURCE POS>   <Mozc POS>
# <SOURCE POS> is converted to <Mozc POS>.

#################################################
# MS-IME
名詞    名詞
形容詞  形容詞
形容動詞        名詞形動
副詞    副詞
連体詞  連体詞
接続詞  接続詞
感動詞  感動詞
接頭語  接頭語
接尾語  接尾一般
さ変名詞        名詞サ変
ざ変名詞        名詞サ変
形動名詞        名詞形動
副詞的名詞      名詞
さ変形動名詞    名詞形動
人名    人名
姓      姓
名      名
地名その他      地名
固有名詞        固有名詞
あわ行五段      動詞ワ行五段
か行五段        動詞カ行五段
が行五段        動詞ガ行五段
さ行五段        動詞サ行五段
た行五段        動詞タ行五段
な行五段        動詞ナ行五段
ば行五段        動詞バ行五段
ま行五段        動詞マ行五段
ら行五段        動詞ラ行五段
一段動詞        動詞一段
姓名接頭語      接頭語
地名接頭語      接頭語
姓名接尾語      接尾人名
地名接尾語      接尾地名
助数詞  助数詞
顔文字  顔文字
その他自立語    名詞
慣用句  独立語
独立語  独立語
短縮よみ 短縮よみ
抑制単語 抑制単語

# Internal POS used in msime.h
さ変形動        名詞形動
地名    地名
国      地名
県      地名
郡      地名
区      地名
市      地名
町      地名
村      地名
駅      地名
社名    組織
組織    組織
建築物  組織
物品    固有名詞
代名詞  名詞
人称代名詞      名詞
指示代名詞      名詞
数      数
数量    数
数詞    数
あわ行  動詞ワ行五段
か行    動詞カ行五段
が行    動詞ガ行五段
さ行    動詞サ行五段
た行    動詞タ行五段
な行    動詞ナ行五段
ば行    動詞バ行五段
ま行    動詞マ行五段
ら行    動詞ラ行五段
あわ行う音便    動詞ワ行五段
か行促音便      動詞カ行五段
ら行変格        動詞ラ変

ハ行四段        動詞ハ行四段

形容詞ｶﾞﾙ       形容詞
形容詞ｹﾞ        形容詞
形容詞ﾒ 形容詞
形容詞ｭｳ        形容詞
形容詞ｳ 形容詞
形容動詞ﾉ       名詞形動
形容動詞ﾀﾙ      名詞形動
形容動詞ｶﾞﾙ     名詞形動
さ変副詞        副詞
副詞ﾆ   副詞
副詞ﾅ   副詞
副詞ﾀﾞ  副詞
副詞ﾄ   副詞
副詞ﾄさ変       副詞
指示連体詞      連体詞
人名接頭語      接頭語
固有接頭語      接頭語
前置助数詞      接頭語
丁寧１接頭語    接頭語
丁寧２接頭語    接頭語
丁寧３接頭語    接頭語
人名接尾語      接尾人名
固有接尾語      接尾一般
社名接尾語      接尾一般
組織接尾語      接尾一般
建築物接尾語    接尾一般
連用接尾語      接尾一般
後置助数詞      助数詞
後置助数詞＋    助数詞
時間助数詞      助数詞
時間助数詞＋    助数詞
丁寧接尾語      接尾一般
丁寧１接尾語    接尾一般
丁寧２接尾語    接尾一般
丁寧３接尾語    接尾一般
丁寧４接尾語    接尾一般
複数接尾語      接尾一般
単漢字  名詞
顔      顔文字
記号    記号
英字    アルファベット
解析不能文字    名詞
短縮語  名詞

#################################################
# ATOK
名詞    名詞
固有人姓        姓
固有人名        名
固有人他        人名
固有地名        地名
固有組織        組織
固有商品        固有名詞
固有一般        固有名詞
名詞サ変        名詞サ変
名詞ザ変        名詞サ変
名詞形動        名詞形動
名サ形動        名詞形動

形容詞          形容詞
形容詞ウ        形容詞
形容詞イ        形容詞
形容詞エ        形容詞
形容動詞        名詞形動
形動タリ        名詞形動

カ行五段        動詞カ行五段
ガ行五段        動詞ガ行五段
サ行五段        動詞サ行五段
タ行五段        動詞タ行五段
ナ行五段        動詞ナ行五段
バ行五段        動詞バ行五段
マ行五段        動詞マ行五段
ラ行五段        動詞ラ行五段
ワ行五段        動詞ワ行五段
ハ行四段        動詞ハ行四段
一段動詞          動詞一段
カ変動詞          動詞カ変
サ変動詞          動詞サ変
ザ変動詞          動詞ザ変
ワ行五段音便      動詞ワ行五段

ラ変動詞  動詞ラ変

# not supported
ナ変動詞
カ行上二段
ガ行上二段
タ行上二段
ダ行上二段
ハ行上二段
バ行上二段
マ行上二段
ヤ行上二段
ラ行上二段
カ行下二段
ガ行下二段
サ行下二段
ザ行下二段
タ行下二段
ダ行下二段
ナ行下二段
ハ行下二段
バ行下二段
マ行下二段
ヤ行下二段
ラ行下二段
ワ行下二段

数詞    数
副詞    副詞
連体詞  連体詞
接続詞  接続詞
感動詞  感動詞
独立語  独立語
接頭語  接頭語
冠数詞  接頭語
接尾語  接尾一般
助数詞  助数詞
単漢字  名詞
終助詞  終助詞
顔文字  顔文字

# ATOK POS ID
# http://support.justsystems.com/faq/1032/app/servlet/qadoc?QID=043885
1    名詞
2    姓
3    名
4    人名
5    地名
6    組織
7    固有名詞
8    固有名詞
9    名詞サ変
10   名詞サ変
11   名詞形動
12   名詞形動

37       形容詞
38       形容詞
68       形容詞
69       形容詞
39       名詞形動
40       名詞形動

23       動詞カ行五段
24       動詞ガ行五段
25       動詞サ行五段
26       動詞タ行五段
27       動詞ナ行五段
28       動詞バ行五段
29       動詞マ行五段
30       動詞ラ行五段
31       動詞ワ行五段
32       動詞ハ行四段
33       動詞一段
34       動詞カ変
35       動詞サ変
36       動詞ザ変
43       動詞ワ行五段

45 動詞ラ変

# not supported
44
46
47
48
49
50
51
52
53
54
55
56
57
58
59
60
61
62
63
64
65
66
67

13  数
14  副詞
15  連体詞
16  接続詞
17  感動詞
18  独立語
19  接頭語
20  接頭語
21  接尾一般
22  助数詞
41  名詞
42  終助詞
70  顔文字

#################################################
# Kotoeri
普通名詞        名詞
人名            人名
地名            地名
固有名詞        固有名詞
その他の固有名詞        固有名詞
サ変名詞                名詞サ変
形容動詞                名詞形動
形容詞                  形容詞

ワ行五段        動詞ワ行五段
カ行五段        動詞カ行五段
サ行五段        動詞サ行五段
タ行五段        動詞タ行五段
ナ行五段        動詞ナ行五段
マ行五段        動詞マ行五段
ラ行五段        動詞ラ行五段
ガ行五段        動詞ガ行五段
バ行五段        動詞バ行五段
一段動詞        動詞一段

サ変動詞          動詞サ変
ザ変動詞          動詞ザ変

連体詞           連体詞
形容動詞         名詞形動
副詞             副詞
接続詞           接続詞
感動詞           感動詞
接頭語           接頭語
人名接尾語       接尾人名
地名接尾語       接尾地名
組織名接尾語     接尾一般
接尾語           接尾一般
数字列接尾語     助数詞
無品詞           独立語
