# boundary.def
#
# boundary.def is a rule file for modeling 'Japanese bunsetsu'.
# User basically expects that his/her input will be handled as 'valid'
# bunsetsu in terms of school grammer. For instance, a bunsetsu starting
# with Japanese particles (はがを) doesn't look valid bunsetsu.
#
# Format
# label feature_regexp penalty.
#
# label: (SUFFIX|PREFIX) suffix rule or prefix rule.
# feature_regexp: POS/feature regexp patterns.
# penalty: penality cost.
#
# basically if a word exists at the beginning or end of user input,
# prefix or suffix rule is applied.

#####################################################
# Manual rules:
# When you find misconversions which are caused by this rule,
# edit the following section.

# Users tend to input verb 基本形 in one segment
SUFFIX 動詞,自立,*,*,*,基本形        0

SUFFIX 名詞,形容動詞語幹,            300

PREFIX 名詞,非自立,副詞可能,*,*,*,際 1000
# want to promote "件"
PREFIX 名詞,接尾,地域,*,*,*,県       1500
# user may input 123回 as 123|回
PREFIX 名詞,接尾,助数詞              1200
PREFIX 名詞,接尾,特殊,               2000
PREFIX 名詞,接尾,一般,               1000
PREFIX 名詞,接尾,サ変接続,           2000
PREFIX 名詞,接尾,                    1000
PREFIX 助詞,格助詞,引用,*,*,*,と     2000
PREFIX 助詞,(格助詞|連体化)          3000
# Suppress し|なし
PREFIX 動詞,自立,*,*,サ変・スル,連用形 3000
PREFIX 動詞,自立,*,*,サ変・スル,*    1000

# general noun might be buried in personal names
# especially we have a lot of first name variants
SUFFIX 名詞,接尾可能,人名            0
SUFFIX 名詞,固有名詞,人名,名         1000
PREFIX 名詞,固有名詞,人名,名         1000
PREFIX 名詞,固有名詞,人名            800

# Suppress なかたに -> な|方|に
# Since ながる is not a natural expression, we can strongly demote it.
# Otherwise this な appears with 方|に.
PREFIX 助動詞,*,*,*,特殊・ダ,体言接続,だ           3000
PREFIX 形容詞,自立,*,*,形容詞・アウオ段,ガル接続,ない 9000

PREFIX 名詞,数,漢数字,               150
PREFIX 名詞,(数|固有名詞|副詞可能|代名詞|一般|サ変接続), 0
# さかれていた => さ|枯れていた
PREFIX 副詞,助詞類接続,*,*,*,*,さ    2000
PREFIX (形容詞,自立|連体詞),   0

SUFFIX 名詞,一般,*,*,*,*,*,気        500

# おうち
SUFFIX 名詞,非自立,副詞可能          1000

# 嘆じた
PREFIX 助動詞,*,*,*,特殊・タ,基本形     3000

SUFFIX 名詞,固有名詞,一般         0
SUFFIX 名詞,接尾,一般 700
SUFFIX 名詞,(数|固有名詞|副詞可能|代名詞|一般|サ変接続|接尾), 300
SUFFIX 連体詞, 0
# These suffixes tend to be input as one word.
SUFFIX 接頭詞,名詞接続,*,*,*,*,*,(各|約|超|当|同|全|元|新|等)  0
# These suffixes tend to be input as one word, but weaker than previous words.
SUFFIX 接頭詞,名詞接続,*,*,*,*,*,(低|未|非|高|大) 1000

# 代用品
# major 接頭辞 will be covered by compound
PREFIX 接頭詞,名詞接続               200

PREFIX 動詞,自立                     100
PREFIX 動詞,非自立                    4000

# http://mozcsuorg.appspot.com/#issue/167
# With this demotion, もふる -> も|フル misconversion is fixed and モフる shows
# up.  However, もおなじです -> も同じです can still be converted.
PREFIX 助詞,係助詞,*,*,*,*,も  2500

# 会いた|がる
SUFFIX 助動詞,*,*,*,特殊・タイ,ガル接続,たい 7000
SUFFIX 接頭詞, 4000

# とっく v.s. 特区
SUFFIX 副詞,一般,	300

# TODO(noriyukit): This is a hack to fix the following conversion problem:
# おぼろづき -> おぼろ|づき  (From compound)
#   lid: 名詞,一般,*,*,*,*,*
#   rid: 動詞,自立,*,*,五段・カ行イ音便,連用形,3
# After rebuilding the dictionary with cl/48606719 that registeres おぼろづき
# (COMPOUND) to manual_filter.def, we should remove this hack.
SUFFIX 動詞,自立,*,*,五段・カ行イ音便,連用形,3	2000

# Suppress incomplete X-接続 forms, e.g.:
#   - かい -> 書い and かい -> 描い   (連用タ接続)
#   - いお -> 言お and ひらこ -> 開こ  (未然ウ接続)
#   - こうふか -> 耕深  (ガル接続)
# See http://mozcsuorg.appspot.com/#issue/187 for more examples.  Putting large
# penalty will be safe because users will input those words in complete forms,
# like かいた.  This rule should be SUFFIX because, if it is PREFIX,
# かいた -> 書い|た is also suppressed.
SUFFIX 動詞,*,*,*,連用タ接続,	7000
SUFFIX 動詞,*,*,*,*,連用タ接続,	7000
SUFFIX 形容詞,*,*,*,*,連用タ接続,	7000
SUFFIX 形容詞,*,*,*,*,ガル接続,	7000
SUFFIX 動詞,自立,*,*,五段・カ行イ音便,未然ウ接続,	7000
SUFFIX 動詞,自立,*,*,五段・ワ行ウ音便,未然ウ接続,	7000

# Manually adjusted POS to prevent some misconversions, e.g.:
#   - わっず -> 割っ+ず  (連用ニ接続,ぬ)
#   - あいの -> 会い+の  (五段・ワ行促音便,丁寧連用形)
#   - ふけ -> ふけ  (仮定形 of ふく, 五段・カ行イ音便)
#   - いとい -> いとい  (連用形 of いとう, 五段・ワ行ウ音便,連用形)
#   - こうふか -> 高付加 vs 交付か  (サ変名詞 + 助詞,か)
# TODO(noriyukit): 仮定形 may be more agressively demoted as users will input
# 仮定形 with particles, like すれ+ば.
PREFIX 動詞,自立,*,*,五段・ワ行促音便,丁寧連用形,*	500
SUFFIX 助動詞,*,*,*,特殊・ヌ,連用ニ接続,ぬ	300
SUFFIX 動詞,自立,*,*,五段・カ行イ音便,仮定形,	3000
SUFFIX 動詞,自立,*,*,五段・カ行イ音便,未然形,	2000
SUFFIX 動詞,自立,*,*,五段・カ行イ音便,連用形,	1000
SUFFIX 動詞,自立,*,*,五段・ワ行ウ音便,連用形,	1000
SUFFIX 動詞,非自立,*,*,五段・カ行イ音便,仮定形,	3000
SUFFIX 動詞,非自立,*,*,五段・カ行イ音便,未然形,	3000
SUFFIX 動詞,非自立,*,*,五段・カ行イ音便,連用形,	1000
SUFFIX 動詞,非自立,*,*,五段・ワ行ウ音便,連用形,	1000
SUFFIX 助詞,副助詞／並立助詞／終助詞,*,*,*,*,か	750

PREFIX 助詞,終助詞,*,*,*,*,ね	2000

# Users wouldn't input words/phrases/sentences ending with 連用形 of いる.
# For instance, かわいい to 河合(noun) + い(連用形 of いる).
SUFFIX 動詞,自立,*,*,一段,連用形,いる  3000

#####################################################
# Automatic rules.
#
# Aotumatic rule is generated as follows:
# prefix: -500 * log(P(begging_of_bunsetsu|POS))
# suffix: -500 * log(P(end_of_bunsetsu|POS))
#
# P(*|POS) is calcurated from the bunsetsu-segmented corpus.

# PREFIX RULE
PREFIX	連体詞,*,*,*,*,*,*	3
PREFIX	接頭詞,数接続,*,*,*,*,*	8
PREFIX	接頭詞,形容詞接続,*,*,*,*,*	9
PREFIX	接頭詞,名詞接続,*,*,*,*,*	5
PREFIX	接頭詞,動詞接続,*,*,*,*,*	0
PREFIX	接続詞,*,*,*,*,*,*	9
PREFIX	感動詞,*,*,*,*,*,*	17
PREFIX	形容詞,非自立,*,*,*,連用テ接続,*	785
PREFIX	形容詞,非自立,*,*,*,連用タ接続,*	65
PREFIX	形容詞,非自立,*,*,*,連用ゴザイ接続,*	0
PREFIX	形容詞,非自立,*,*,*,未然ウ接続,*	0
PREFIX	形容詞,非自立,*,*,*,文語基本形,*	52
PREFIX	形容詞,非自立,*,*,*,基本形,*	292
PREFIX	形容詞,非自立,*,*,*,体言接続,*	346
PREFIX	形容詞,非自立,*,*,*,仮定形,*	0
PREFIX	形容詞,非自立,*,*,*,ガル接続,*	836
PREFIX	形容詞,自立,*,*,*,連用テ接続,*	22
PREFIX	形容詞,自立,*,*,*,連用タ接続,*	42
PREFIX	形容詞,自立,*,*,*,連用ゴザイ接続,*	32
PREFIX	形容詞,自立,*,*,*,未然ヌ接続,*	0
PREFIX	形容詞,自立,*,*,*,未然ウ接続,*	0
PREFIX	形容詞,自立,*,*,*,文語基本形,*	76
PREFIX	形容詞,自立,*,*,*,基本形-促音便,*	0
PREFIX	形容詞,自立,*,*,*,基本形,*	45
PREFIX	形容詞,自立,*,*,*,命令ｅ,*	143
PREFIX	形容詞,自立,*,*,*,体言接続,*	58
PREFIX	形容詞,自立,*,*,*,仮定縮約２,*	0
PREFIX	形容詞,自立,*,*,*,仮定形,*	42
PREFIX	形容詞,自立,*,*,*,ガル接続,*	19
PREFIX	形容詞,接尾,*,*,*,連用テ接続,*	1213
PREFIX	形容詞,接尾,*,*,*,連用タ接続,*	5000
PREFIX	形容詞,接尾,*,*,*,基本形,*	998
PREFIX	形容詞,接尾,*,*,*,ガル接続,*	184
PREFIX	名詞,非自立,形容動詞語幹,*,*,*,*	648
PREFIX	名詞,非自立,助動詞語幹,*,*,*,*	1504
PREFIX	名詞,非自立,副詞可能,*,*,*,*	92
PREFIX	名詞,非自立,一般,*,*,*,*	355
PREFIX	名詞,特殊,助動詞語幹,*,*,*,*	946
PREFIX	名詞,数,*,*,*,*,*	208
PREFIX	名詞,接続詞的,*,*,*,*,*	8
PREFIX	名詞,接尾,特殊,*,*,*,*	1264
PREFIX	名詞,接尾,形容動詞語幹,*,*,*,*	829
PREFIX	名詞,接尾,地域,*,*,*,*	494
PREFIX	名詞,接尾,助数詞,*,*,*,*	157
PREFIX	名詞,接尾,助動詞語幹,*,*,*,*	1459
PREFIX	名詞,接尾,副詞可能,*,*,*,*	322
PREFIX	名詞,接尾,人名,*,*,*,*	34
PREFIX	名詞,接尾,一般,*,*,*,*	331
PREFIX	名詞,接尾,サ変接続,*,*,*,*	517
PREFIX	名詞,形容動詞語幹,*,*,*,*,*	7
PREFIX	名詞,固有名詞,組織,*,*,*,*	20
PREFIX	名詞,固有名詞,地域,国,*,*,*	13
PREFIX	名詞,固有名詞,地域,一般,*,*,*	6
PREFIX	名詞,固有名詞,人名,姓,*,*,*	2
PREFIX	名詞,固有名詞,人名,名,*,*,*	27
PREFIX	名詞,固有名詞,人名,一般,*,*,*	33
PREFIX	名詞,固有名詞,一般,*,*,*,*	28
PREFIX	名詞,動詞非自立的,*,*,*,*,*	46
PREFIX	名詞,副詞可能,*,*,*,*,*	13
PREFIX	名詞,代名詞,縮約,*,*,*,*	58
PREFIX	名詞,代名詞,一般,*,*,*,*	4
PREFIX	名詞,一般,*,*,*,*,*	26
PREFIX	名詞,ナイ形容詞語幹,*,*,*,*,*	5
PREFIX	名詞,サ変接続,*,*,*,*,*	39
PREFIX	動詞,非自立,*,*,*,連用形,*	1059
PREFIX	動詞,非自立,*,*,*,連用タ接続,*	540
PREFIX	動詞,非自立,*,*,*,未然特殊,*	1039
PREFIX	動詞,非自立,*,*,*,未然形,*	835
PREFIX	動詞,非自立,*,*,*,未然ウ接続,*	1051
PREFIX	動詞,非自立,*,*,*,基本形,*	1403
PREFIX	動詞,非自立,*,*,*,命令ｙｏ,*	77
PREFIX	動詞,非自立,*,*,*,命令ｒｏ,*	381
PREFIX	動詞,非自立,*,*,*,命令ｉ,*	21
PREFIX	動詞,非自立,*,*,*,命令ｅ,*	981
PREFIX	動詞,非自立,*,*,*,体言接続特殊２,*	1301
PREFIX	動詞,非自立,*,*,*,体言接続特殊,*	752
PREFIX	動詞,非自立,*,*,*,仮定縮約１,*	0
PREFIX	動詞,非自立,*,*,*,仮定形,*	944
PREFIX	動詞,自立,*,*,*,連用形,*	192
PREFIX	動詞,自立,*,*,*,連用タ接続,*	44
PREFIX	動詞,自立,*,*,*,未然特殊,*	25
PREFIX	動詞,自立,*,*,*,未然形,*	58
PREFIX	動詞,自立,*,*,*,未然レル接続,*	1251
PREFIX	動詞,自立,*,*,*,未然ヌ接続,*	446
PREFIX	動詞,自立,*,*,*,未然ウ接続,*	135
PREFIX	動詞,自立,*,*,*,文語基本形,*	234
PREFIX	動詞,自立,*,*,*,基本形,*	115
PREFIX	動詞,自立,*,*,*,命令ｙｏ,*	162
PREFIX	動詞,自立,*,*,*,命令ｒｏ,*	187
PREFIX	動詞,自立,*,*,*,命令ｉ,*	0
PREFIX	動詞,自立,*,*,*,命令ｅ,*	74
PREFIX	動詞,自立,*,*,*,体言接続特殊２,*	208
PREFIX	動詞,自立,*,*,*,体言接続特殊,*	168
PREFIX	動詞,自立,*,*,*,仮定縮約１,*	0
PREFIX	動詞,自立,*,*,*,仮定形,*	105
PREFIX	動詞,接尾,*,*,*,連用形,*	2194
PREFIX	動詞,接尾,*,*,*,連用タ接続,*	509
PREFIX	動詞,接尾,*,*,*,未然形,*	1687
PREFIX	動詞,接尾,*,*,*,未然ウ接続,*	5000
PREFIX	動詞,接尾,*,*,*,基本形,*	2499
PREFIX	動詞,接尾,*,*,*,命令ｙｏ,*	5000
PREFIX	動詞,接尾,*,*,*,命令ｒｏ,*	5000
PREFIX	動詞,接尾,*,*,*,体言接続特殊,*	5000
PREFIX	動詞,接尾,*,*,*,仮定縮約１,*	66
PREFIX	動詞,接尾,*,*,*,仮定形,*	1510
PREFIX	助詞,連体化,*,*,*,*,*	1482
PREFIX	助詞,終助詞,*,*,*,*,*	1322
PREFIX	助詞,特殊,*,*,*,*,*	125
PREFIX	助詞,格助詞,引用,*,*,*,*	686
PREFIX	助詞,格助詞,一般,*,*,*,*	1559
PREFIX	助詞,接続助詞,*,*,*,*,*	2510
PREFIX	助詞,副詞化,*,*,*,*,*	2896
PREFIX	助詞,副助詞／並立助詞／終助詞,*,*,*,*,*	2006
PREFIX	助詞,副助詞,*,*,*,*,*	1306
PREFIX	助詞,係助詞,*,*,*,*,*	1749
PREFIX	助詞,並立助詞,*,*,*,*,*	1346
PREFIX	助動詞,*,*,*,*,音便基本形,*	114
PREFIX	助動詞,*,*,*,*,連用形,*	1490
PREFIX	助動詞,*,*,*,*,連用ニ接続,*	2582
PREFIX	助動詞,*,*,*,*,連用デ接続,*	1702
PREFIX	助動詞,*,*,*,*,連用テ接続,*	1627
PREFIX	助動詞,*,*,*,*,連用タ接続,*	1601
PREFIX	助動詞,*,*,*,*,未然形,*	2196
PREFIX	助動詞,*,*,*,*,未然ヌ接続,*	0
PREFIX	助動詞,*,*,*,*,未然ウ接続,*	1950
PREFIX	助動詞,*,*,*,*,文語基本形,*	309
PREFIX	助動詞,*,*,*,*,基本形,*	2001
PREFIX	助動詞,*,*,*,*,命令ｉ,*	143
PREFIX	助動詞,*,*,*,*,命令ｅ,*	749
PREFIX	助動詞,*,*,*,*,体言接続特殊,*	5000
PREFIX	助動詞,*,*,*,*,体言接続,*	1362
PREFIX	助動詞,*,*,*,*,仮定縮約２,*	5000
PREFIX	助動詞,*,*,*,*,仮定形,*	2197
PREFIX	助動詞,*,*,*,*,ガル接続,*	649
PREFIX	副詞,助詞類接続,*,*,*,*,*	12
PREFIX	副詞,一般,*,*,*,*,*	11
PREFIX	フィラー,*,*,*,*,*,*	113
PREFIX	その他,間投,*,*,*,*,*	0

# SUFFIX RULE
SUFFIX	連体詞,*,*,*,*,*,*	51
SUFFIX	接頭詞,数接続,*,*,*,*,*	158
SUFFIX	接頭詞,形容詞接続,*,*,*,*,*	118
SUFFIX	接頭詞,名詞接続,*,*,*,*,*	580
SUFFIX	接頭詞,動詞接続,*,*,*,*,*	1039
SUFFIX	接続詞,*,*,*,*,*,*	16
SUFFIX	感動詞,*,*,*,*,*,*	124
SUFFIX	形容詞,非自立,*,*,*,連用テ接続,*	342
SUFFIX	形容詞,非自立,*,*,*,連用タ接続,*	5000
SUFFIX	形容詞,非自立,*,*,*,連用ゴザイ接続,*	549
SUFFIX	形容詞,非自立,*,*,*,未然ウ接続,*	0
SUFFIX	形容詞,非自立,*,*,*,文語基本形,*	255
SUFFIX	形容詞,非自立,*,*,*,基本形,*	545
SUFFIX	形容詞,非自立,*,*,*,体言接続,*	0
SUFFIX	形容詞,非自立,*,*,*,仮定形,*	5000
SUFFIX	形容詞,非自立,*,*,*,ガル接続,*	836
SUFFIX	形容詞,自立,*,*,*,連用テ接続,*	289
SUFFIX	形容詞,自立,*,*,*,連用タ接続,*	2855
SUFFIX	形容詞,自立,*,*,*,連用ゴザイ接続,*	116
SUFFIX	形容詞,自立,*,*,*,未然ヌ接続,*	279
SUFFIX	形容詞,自立,*,*,*,未然ウ接続,*	5000
SUFFIX	形容詞,自立,*,*,*,文語基本形,*	369
SUFFIX	形容詞,自立,*,*,*,基本形-促音便,*	5000
SUFFIX	形容詞,自立,*,*,*,基本形,*	289
SUFFIX	形容詞,自立,*,*,*,命令ｅ,*	0
SUFFIX	形容詞,自立,*,*,*,体言接続,*	92
SUFFIX	形容詞,自立,*,*,*,仮定縮約２,*	0
SUFFIX	形容詞,自立,*,*,*,仮定形,*	5000
SUFFIX	形容詞,自立,*,*,*,ガル接続,*	1037
SUFFIX	形容詞,接尾,*,*,*,連用テ接続,*	346
SUFFIX	形容詞,接尾,*,*,*,連用タ接続,*	5000
SUFFIX	形容詞,接尾,*,*,*,基本形,*	210
SUFFIX	形容詞,接尾,*,*,*,ガル接続,*	626
SUFFIX	名詞,非自立,形容動詞語幹,*,*,*,*	1032
SUFFIX	名詞,非自立,助動詞語幹,*,*,*,*	1497
SUFFIX	名詞,非自立,副詞可能,*,*,*,*	621
SUFFIX	名詞,非自立,一般,*,*,*,*	1315
SUFFIX	名詞,特殊,助動詞語幹,*,*,*,*	2115
SUFFIX	名詞,数,*,*,*,*,*	307
SUFFIX	名詞,接続詞的,*,*,*,*,*	289
SUFFIX	名詞,接尾,特殊,*,*,*,*	814
SUFFIX	名詞,接尾,形容動詞語幹,*,*,*,*	946
SUFFIX	名詞,接尾,地域,*,*,*,*	288
SUFFIX	名詞,接尾,助数詞,*,*,*,*	368
SUFFIX	名詞,接尾,助動詞語幹,*,*,*,*	941
SUFFIX	名詞,接尾,副詞可能,*,*,*,*	530
SUFFIX	名詞,接尾,人名,*,*,*,*	579
SUFFIX	名詞,接尾,一般,*,*,*,*	501
SUFFIX	名詞,接尾,サ変接続,*,*,*,*	731
SUFFIX	名詞,形容動詞語幹,*,*,*,*,*	826
SUFFIX	名詞,固有名詞,組織,*,*,*,*	129
SUFFIX	名詞,固有名詞,地域,国,*,*,*	431
SUFFIX	名詞,固有名詞,地域,一般,*,*,*	313
SUFFIX	名詞,固有名詞,人名,姓,*,*,*	178
SUFFIX	名詞,固有名詞,人名,名,*,*,*	184
SUFFIX	名詞,固有名詞,人名,一般,*,*,*	266
SUFFIX	名詞,固有名詞,一般,*,*,*,*	201
SUFFIX	名詞,動詞非自立的,*,*,*,*,*	97
SUFFIX	名詞,副詞可能,*,*,*,*,*	498
SUFFIX	名詞,代名詞,縮約,*,*,*,*	125
SUFFIX	名詞,代名詞,一般,*,*,*,*	1033
SUFFIX	名詞,一般,*,*,*,*,*	463
SUFFIX	名詞,ナイ形容詞語幹,*,*,*,*,*	919
SUFFIX	名詞,サ変接続,*,*,*,*,*	384
SUFFIX	動詞,非自立,*,*,*,連用形,*	1419
SUFFIX	動詞,非自立,*,*,*,連用タ接続,*	1521
SUFFIX	動詞,非自立,*,*,*,未然特殊,*	346
SUFFIX	動詞,非自立,*,*,*,未然形,*	2050
SUFFIX	動詞,非自立,*,*,*,未然ウ接続,*	2815
SUFFIX	動詞,非自立,*,*,*,基本形,*	317
SUFFIX	動詞,非自立,*,*,*,命令ｙｏ,*	0
SUFFIX	動詞,非自立,*,*,*,命令ｒｏ,*	71
SUFFIX	動詞,非自立,*,*,*,命令ｉ,*	60
SUFFIX	動詞,非自立,*,*,*,命令ｅ,*	258
SUFFIX	動詞,非自立,*,*,*,体言接続特殊２,*	448
SUFFIX	動詞,非自立,*,*,*,体言接続特殊,*	405
SUFFIX	動詞,非自立,*,*,*,仮定縮約１,*	265
SUFFIX	動詞,非自立,*,*,*,仮定形,*	2143
SUFFIX	動詞,自立,*,*,*,連用形,*	1026
SUFFIX	動詞,自立,*,*,*,連用タ接続,*	2378
SUFFIX	動詞,自立,*,*,*,未然特殊,*	773
SUFFIX	動詞,自立,*,*,*,未然形,*	2628
SUFFIX	動詞,自立,*,*,*,未然レル接続,*	3321
SUFFIX	動詞,自立,*,*,*,未然ヌ接続,*	1404
SUFFIX	動詞,自立,*,*,*,未然ウ接続,*	2296
SUFFIX	動詞,自立,*,*,*,文語基本形,*	492
SUFFIX	動詞,自立,*,*,*,基本形,*	244
SUFFIX	動詞,自立,*,*,*,命令ｙｏ,*	109
SUFFIX	動詞,自立,*,*,*,命令ｒｏ,*	180
SUFFIX	動詞,自立,*,*,*,命令ｉ,*	23
SUFFIX	動詞,自立,*,*,*,命令ｅ,*	128
SUFFIX	動詞,自立,*,*,*,体言接続特殊２,*	297
SUFFIX	動詞,自立,*,*,*,体言接続特殊,*	448
SUFFIX	動詞,自立,*,*,*,仮定縮約１,*	346
SUFFIX	動詞,自立,*,*,*,仮定形,*	2349
SUFFIX	動詞,接尾,*,*,*,連用形,*	1178
SUFFIX	動詞,接尾,*,*,*,連用タ接続,*	576
SUFFIX	動詞,接尾,*,*,*,未然形,*	2289
SUFFIX	動詞,接尾,*,*,*,未然ウ接続,*	5000
SUFFIX	動詞,接尾,*,*,*,基本形,*	214
SUFFIX	動詞,接尾,*,*,*,命令ｙｏ,*	0
SUFFIX	動詞,接尾,*,*,*,命令ｒｏ,*	346
SUFFIX	動詞,接尾,*,*,*,体言接続特殊,*	0
SUFFIX	動詞,接尾,*,*,*,仮定縮約１,*	5000
SUFFIX	動詞,接尾,*,*,*,仮定形,*	1856
SUFFIX	助詞,連体化,*,*,*,*,*	8
SUFFIX	助詞,終助詞,*,*,*,*,*	91
SUFFIX	助詞,特殊,*,*,*,*,*	92
SUFFIX	助詞,格助詞,引用,*,*,*,*	50
SUFFIX	助詞,格助詞,一般,*,*,*,*	52
SUFFIX	助詞,接続助詞,*,*,*,*,*	248
SUFFIX	助詞,副詞化,*,*,*,*,*	11
SUFFIX	助詞,副助詞／並立助詞／終助詞,*,*,*,*,*	209
SUFFIX	助詞,副助詞,*,*,*,*,*	301
SUFFIX	助詞,係助詞,*,*,*,*,*	27
SUFFIX	助詞,並立助詞,*,*,*,*,*	48
SUFFIX	助動詞,*,*,*,*,音便基本形,*	392
SUFFIX	助動詞,*,*,*,*,連用形,*	955
SUFFIX	助動詞,*,*,*,*,連用ニ接続,*	207
SUFFIX	助動詞,*,*,*,*,連用デ接続,*	2126
SUFFIX	助動詞,*,*,*,*,連用テ接続,*	469
SUFFIX	助動詞,*,*,*,*,連用タ接続,*	2708
SUFFIX	助動詞,*,*,*,*,未然形,*	1357
SUFFIX	助動詞,*,*,*,*,未然ヌ接続,*	0
SUFFIX	助動詞,*,*,*,*,未然ウ接続,*	1603
SUFFIX	助動詞,*,*,*,*,文語基本形,*	656
SUFFIX	助動詞,*,*,*,*,基本形,*	259
SUFFIX	助動詞,*,*,*,*,命令ｉ,*	0
SUFFIX	助動詞,*,*,*,*,命令ｅ,*	55
SUFFIX	助動詞,*,*,*,*,体言接続特殊,*	0
SUFFIX	助動詞,*,*,*,*,体言接続,*	175
SUFFIX	助動詞,*,*,*,*,仮定縮約２,*	37
SUFFIX	助動詞,*,*,*,*,仮定形,*	106
SUFFIX	助動詞,*,*,*,*,ガル接続,*	528
SUFFIX	副詞,助詞類接続,*,*,*,*,*	163
SUFFIX	副詞,一般,*,*,*,*,*	51
SUFFIX	フィラー,*,*,*,*,*,*	206
SUFFIX	その他,間投,*,*,*,*,*	549
