// Microsoft Visual C++ generated resource script.
//
#include "server\mozc_cache_service_resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include "afxres.h"

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// Japanese resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_JPN)
#ifdef _WIN32
LANGUAGE LANG_JAPANESE, SUBLANG_DEFAULT
#endif //_WIN32


/////////////////////////////////////////////////////////////////////////////
//
// Version
//

#ifndef MOZC_RES_INTERNAL_NAME
  #if defined(GOOGLE_JAPANESE_INPUT_BUILD)
    #define MOZC_RES_INTERNAL_NAME "GoogleIMEJaCacheService"
  #elif defined(MOZC_BUILD)
    #define MOZC_RES_INTERNAL_NAME "mozc_cache_service"
  #endif
#endif
#ifndef MOZC_RES_ORIGINAL_FILENAME
  #if defined(GOOGLE_JAPANESE_INPUT_BUILD)
    #define MOZC_RES_ORIGINAL_FILENAME "GoogleIMEJaCacheService.exe"
  #elif defined(MOZC_BUILD)
    #define MOZC_RES_ORIGINAL_FILENAME "mozc_cache_service.exe"
  #endif
#endif
#ifndef MOZC_RES_FILE_DESCRIPTION
  #if defined(GOOGLE_JAPANESE_INPUT_BUILD)
    #define MOZC_RES_FILE_DESCRIPTION "Google 日本語入力 キャッシュサービス"
  #elif defined(MOZC_BUILD)
    #define MOZC_RES_FILE_DESCRIPTION "Mozc Cache Service"
  #endif
#endif
#ifndef MOZC_RES_FILETYPE
  #define MOZC_RES_FILETYPE VFT_APP
#endif

#ifdef MOZC_RES_USE_TEMPLATE
  #include "build_tools/mozc_win32_resource_template.rc"
#endif    // MOZC_RES_USE_TEMPLATE


#endif    // Japanese resources

/////////////////////////////////////////////////////////////////////////////
//
// String Table
//
STRINGTABLE LANGUAGE LANG_JAPANESE, SUBLANG_DEFAULT
BEGIN
    // It is highly recommended not to translate the display name. (http://b/issue?id=2362557)
    // This helps users who want to control the service via "net" command from the command prompt and/or batch files.
#if defined(GOOGLE_JAPANESE_INPUT_BUILD)
    IDS_DISPLAYNAME         "Google Japanese Input Cache Service"
    IDS_DESCRIPTION         "Google 日本語入力 の辞書を常に物理メモリに配置し、変換を高速化します。停止しても日本語入力はできますが、パフォーマンスが低下する場合があります。"
#elif defined(MOZC_BUILD)
    IDS_DISPLAYNAME         "Mozc Cache Service"
    IDS_DESCRIPTION         "Mozc の辞書を常に物理メモリに配置し、変換を高速化します。停止しても日本語入力はできますが、パフォーマンスが低下する場合があります。"
#endif
END

STRINGTABLE LANGUAGE LANG_ENGLISH, SUBLANG_NEUTRAL
BEGIN
#if defined(GOOGLE_JAPANESE_INPUT_BUILD)
    IDS_DISPLAYNAME         "Google Japanese Input Cache Service"
    IDS_DESCRIPTION         "Always allocates conversion dictionary of Google Japanese input into physical memory."
#elif defined(MOZC_BUILD)
    IDS_DISPLAYNAME         "Mozc Cache Service"
    IDS_DESCRIPTION         "Always allocates conversion dictionary of Mozc into physical memory."
#endif
END

#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//


/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

