# Grouping rule for UserHistroyLearning
# LHS POS is handled as RHS group. If two POSes have the same
# group, they have the same POS group.
# This is used for defining "Replaceable" condition.

その他     その他
フィラー   その他

感動詞     感動詞

記号   記号

形容詞,自立   形容詞,自立
形容詞,接尾   形容詞,接尾
形容詞,非自立  形容詞,非自立

助詞  助詞
助動詞 助動詞

接続詞  接続詞

動詞,自立   動詞,自立
動詞,接尾   動詞,接尾
動詞,非自立 動詞,非自立

副詞  副詞

名詞,サ変接続       名詞
名詞,ナイ形容詞語幹 名詞
名詞,一般           名詞
名詞,引用文字列     名詞
名詞,形容動詞語幹   名詞
名詞,固有名詞,一般         固有名詞
名詞,固有名詞,人名,一般   人名
名詞,固有名詞,人名,姓     人名
名詞,固有名詞,人名,名     人名
名詞,固有名詞,組織,*      人名
名詞,固有名詞,地域,一般   固有名詞
名詞,固有名詞,地域,国     固有名詞
名詞,数                   数
名詞,接続詞的             名詞
名詞,接尾,サ変接続        接尾
名詞,接尾,一般            接尾
名詞,接尾,形容動詞語幹    接尾
名詞,接尾,助数詞          接尾
名詞,接尾,助動詞語幹      接尾
名詞,接尾,人名      接尾
名詞,接尾,地域      接尾
名詞,接尾,特殊      接尾
名詞,接尾,副詞可能  接尾
名詞,代名詞,一般    名詞
名詞,代名詞,縮約    名詞
名詞,非自立         名詞,非自立
名詞,副詞可能       名詞
名詞                名詞,その他

連体詞 連体詞

# default rule
* その他
