// Copyright 2010-2020, Google Inc.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
// copyright notice, this list of conditions and the following disclaimer
// in the documentation and/or other materials provided with the
// distribution.
//     * Neither the name of Google Inc. nor the names of its
// contributors may be used to endorse or promote products derived from
// this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "base/status.h"

namespace mozc {

std::ostream &operator<<(std::ostream &os, const Status &status) {
  switch (status.code()) {
    case StatusCode::kOk:
      return os << "OK";
    case StatusCode::kCancelled:
      os << "CANCELLED: ";
      break;
    case StatusCode::kUnknown:
      os << "UNKNOWN: ";
      break;
    case StatusCode::kInvalidArgument:
      os << "INVALID_ARGUMENT: ";
      break;
    case StatusCode::kDeadlineExceeded:
      os << "DEADLINE_EXCEEDED: ";
      break;
    case StatusCode::kNotFound:
      os << "NOT_FOUND: ";
      break;
    case StatusCode::kAlreadyExists:
      os << "ALREADY_EXISTS: ";
      break;
    case StatusCode::kPermissionDenied:
      os << "PERMISSION_DENIED: ";
      break;
    case StatusCode::kResourceExhausted:
      os << "RESOURCE_EXHAUSTED: ";
      break;
    case StatusCode::kFailedPrecondition:
      os << "FAILED_PRECONDITION: ";
      break;
    case StatusCode::kAborted:
      os << "ABORTED: ";
      break;
    case StatusCode::kOutOfRange:
      os << "OUT_OF_RANGE: ";
      break;
    case StatusCode::kUnimplemented:
      os << "UNIMPLEMENTED: ";
      break;
    default:
      os << "UNDEFINED_CODE(" << static_cast<int>(status.code()) << "): ";
      break;
  }
  return os << status.message();
}

}  // namespace mozc
