// Copyright 2010-2020, Google Inc.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
// copyright notice, this list of conditions and the following disclaimer
// in the documentation and/or other materials provided with the
// distribution.
//     * Neither the name of Google Inc. nor the names of its
// contributors may be used to endorse or promote products derived from
// this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// This file is for Mac used in mozc_tool_libmain.cc

#include <memory>

#include "base/flags.h"
#ifdef __APPLE__
#include "base/mac_util.h"
#endif  // __APPLE__
#include "client/client_interface.h"
#include "renderer/renderer_client.h"

#ifdef __APPLE__
DEFINE_bool(register_prelauncher, false, "Register prelauncher to login item.");
#endif  // __APPLE__

int RunPrelaunchProcesses(int argc, char *argv[]) {
#ifdef __APPLE__
  if (FLAGS_register_prelauncher) {
    mozc::MacUtil::AddPrelauncherLoginItem();
  }
#endif  // __APPLE__

  {
    std::unique_ptr<mozc::client::ClientInterface> converter_client(
        mozc::client::ClientFactory::NewClient());
    converter_client->set_suppress_error_dialog(true);
    converter_client->EnsureConnection();
  }

  {
    std::unique_ptr<mozc::renderer::RendererClient> renderer_client(
        new mozc::renderer::RendererClient);
    renderer_client->set_suppress_error_dialog(true);
    renderer_client->Activate();
  }
  return 0;
}
