// Copyright 2010-2020, Google Inc.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
// copyright notice, this list of conditions and the following disclaimer
// in the documentation and/or other materials provided with the
// distribution.
//     * Neither the name of Google Inc. nor the names of its
// contributors may be used to endorse or promote products derived from
// this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef MOZC_GUI_DICTIONARY_TOOL_FIND_DIALOG_H_
#define MOZC_GUI_DICTIONARY_TOOL_FIND_DIALOG_H_

#include <QtWidgets/QDialog>

#include "gui/dictionary_tool/ui_find_dialog.h"

class QTableWidget;
class QTableWidgetItem;

namespace mozc {
namespace gui {

class FindDialog : public QDialog, private Ui::FindDialog {
  Q_OBJECT

 public:
  FindDialog(QWidget *parent, QTableWidget *table);
  ~FindDialog() override;

 protected:
  void closeEvent(QCloseEvent *event) override;
  void showEvent(QShowEvent *event) override;

 private slots:
  void FindForward();
  void FindBackward();
  void UpdateUIStatus();
  void LineEditChanged(const QString &str);

 private:
  enum Direction { FORWARD, BACKWARD };

  bool Match(const QString &query, int row, int column);
  void Find(Direction direction);

  QTableWidget *table_;
  QTableWidgetItem *last_item_;
};
}  // namespace gui
}  // namespace mozc

#endif  // MOZC_GUI_DICTIONARY_TOOL_FIND_DIALOG_H_
