// Copyright 2010-2020, Google Inc.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
// copyright notice, this list of conditions and the following disclaimer
// in the documentation and/or other materials provided with the
// distribution.
//     * Neither the name of Google Inc. nor the names of its
// contributors may be used to endorse or promote products derived from
// this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "engine/user_data_manager_mock.h"

#include <string>

namespace mozc {

UserDataManagerMock::UserDataManagerMock() {}

UserDataManagerMock::~UserDataManagerMock() {}

bool UserDataManagerMock::Sync() {
  function_counters_["Sync"]++;
  return true;
}

bool UserDataManagerMock::Reload() {
  function_counters_["Reload"]++;
  return true;
}

bool UserDataManagerMock::ClearUserHistory() {
  function_counters_["ClearUserHistory"]++;
  return true;
}

bool UserDataManagerMock::ClearUserPrediction() {
  function_counters_["ClearUserPrediction"]++;
  return true;
}

bool UserDataManagerMock::ClearUnusedUserPrediction() {
  function_counters_["ClearUnusedUserPrediction"]++;
  return true;
}

bool UserDataManagerMock::ClearUserPredictionEntry(const std::string &key,
                                                   const std::string &value) {
  last_cleared_key_ = key;
  last_cleared_value_ = value;
  function_counters_["ClearUserPredictionEntry"]++;
  return true;
}

bool UserDataManagerMock::Wait() {
  function_counters_["Wait"]++;
  return true;
}

int UserDataManagerMock::GetFunctionCallCount(const std::string &name) const {
  auto iter = function_counters_.find(name);
  return (iter == function_counters_.end()) ? 0 : iter->second;
}

const std::string &UserDataManagerMock::GetLastClearedKey() const {
  return last_cleared_key_;
}

const std::string &UserDataManagerMock::GetLastClearedValue() const {
  return last_cleared_value_;
}

}  // namespace mozc
