// Copyright 2010-2020, Google Inc.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
// copyright notice, this list of conditions and the following disclaimer
// in the documentation and/or other materials provided with the
// distribution.
//     * Neither the name of Google Inc. nor the names of its
// contributors may be used to endorse or promote products derived from
// this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#import "DialogsController.h"

#include <stdlib.h>

#import "Uninstaller.h"

enum ModalResult {
  DIALOG_OK,
  DIALOG_CANCEL,
};

@implementation DialogsController

- (void)awakeFromNib {
  [confirmDialog_ center];
}

- (BOOL)waitForModalWindow:(NSWindow *)window {
  [window center];
  uint result = [[NSApplication sharedApplication] runModalForWindow:window];
  [window orderOut:nil];
  return result != DIALOG_CANCEL;
}

- (BOOL)reportAuthError {
  return [self waitForModalWindow:reportAuthErrorDialog_];
}

- (BOOL)reportUninstallSuccess {
  return [self waitForModalWindow:uninstallSucceededDialog_];
}

- (BOOL)reportUninstallError {
  return [self waitForModalWindow:uninstallErrorDialog_];
}

- (IBAction)cancel:(id)sender {
  [[NSApplication sharedApplication] terminate:self];
}

- (IBAction)runUninstall:(id)sender {
  [confirmDialog_ orderOut:nil];
  [Uninstaller doUninstall:self];
  [[NSApplication sharedApplication] terminate:self];
}

- (IBAction)dialogOK:(id)sender {
  [[NSApplication sharedApplication] stopModalWithCode:DIALOG_OK];
}

- (IBAction)dialogCancel:(id)sender {
  [[NSApplication sharedApplication] stopModalWithCode:DIALOG_CANCEL];
}

@end
